/**
 * Copyright 2019-2021 Britanicus <marcusbritanicus@gmail.com>
 *
 * This file is a part of DesQ Shell (https://gitlab.com/DesQ/Shell)
 * DesQ Shell is the main Shell UI of the DesQ Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * at your option, any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 **/

#include "Views.hpp"
#include <desq/FSWatcher.hpp>

DesQCategoryView::DesQCategoryView( QWidget *parent ) : QListWidget( parent ) {
    setViewMode( QListView::ListMode );
    setIconSize( QSize( 32, 32 ) );
    setGridSize( QSize( 230, 36 ) );
    setFixedWidth( 250 );
    setFrameStyle( QFrame::NoFrame );
    viewport()->setAutoFillBackground( false );
    setVerticalScrollBarPolicy( Qt::ScrollBarAlwaysOff );
    setHorizontalScrollBarPolicy( Qt::ScrollBarAlwaysOff );

    QStringList categories = {
        "All Applications", "Accessories", "Development", "Education",      "Games",    "Graphics",
        "Internet",         "Multimedia",  "Office",      "Science & Math", "Settings", "System", "Wine"
    };

    QStringList categoryIcons = {
        "preferences-system-windows", "applications-utilities",  "applications-development", "applications-education", "applications-games", "applications-graphics",
        "applications-internet",      "applications-multimedia", "applications-office",      "applications-science",   "preferences-other",  "applications-system",
        "wine"
    };

    for ( int i = 0; i < categories.count(); i++ ) {
        QListWidgetItem *item = new QListWidgetItem( QIcon::fromTheme( categoryIcons.value( i ) ), categories.at( i ) );
        item->setSizeHint( QSize( 230, 36 ) );
        item->setTextAlignment( Qt::AlignVCenter | Qt::AlignLeft );

        addItem( item );
    }

    connect(
        this, &QListWidget::itemClicked, this, [ = ] ( QListWidgetItem *item ) {
            emit loadCategory( item->text() );
        }
    );

    connect(
        this, &QListWidget::itemActivated, this, [ = ] ( QListWidgetItem *item ) {
            emit loadCategory( item->text() );
        }
    );

    setCurrentItem( item( 0 ) );
}


void DesQCategoryView::setCategories( QStringList categories ) {
    if ( not categories.contains( "All Applications" ) ) {
        categories << "All Applications";
    }

    for ( int r = 0; r < 13; r++ ) {
        QListWidgetItem *itm = item( r );

        if ( not categories.contains( itm->text() ) ) {
            itm->setHidden( true );
        }

        else {
            itm->setHidden( false );
        }
    }
}


void DesQCategoryView::mouseReleaseEvent( QMouseEvent *mEvent ) {
    if ( not itemAt( mEvent->pos() ) ) {
        emit closeMenu();
    }

    QListWidget::mouseReleaseEvent( mEvent );
}


void DesQCategoryView::keyPressEvent( QKeyEvent *kEvent ) {
    if ( kEvent->key() == Qt::Key_Slash ) {
        emit focusSearch();
        kEvent->accept();

        return;
    }

    QListWidget::keyPressEvent( kEvent );
}


DesQApplicationsView::DesQApplicationsView( QWidget *parent ) : QListView( parent ) {
    setViewMode( QListView::IconMode );
    setFlow( QListView::LeftToRight );
    setWrapping( true );
    setWordWrap( false );
    setTextElideMode( Qt::ElideRight );
    setResizeMode( QListView::Adjust );
    setMovement( QListView::Static );
    setIconSize( QSize( 64, 64 ) );
    setGridSize( QSize( 144, 108 ) );
    setFrameStyle( QFrame::NoFrame );
    setUniformItemSizes( true );
    viewport()->setAutoFillBackground( false );

    setItemDelegate( new DesQAppsViewDelegate() );

    appsModel  = new QStandardItemModel( this );
    proxyModel = new QSortFilterProxyModel( this );
    proxyModel->setFilterCaseSensitivity( Qt::CaseInsensitive );
    proxyModel->setFilterRole( Qt::UserRole + 1 );

    proxyModel->setSourceModel( appsModel );
    setModel( proxyModel );

    parseDesktops();
    loadCategory( "All Applications" );

    DesQ::FSWatcher *fsw = new DesQ::FSWatcher();
    reloadTimer = new QBasicTimer();

    connect(
        fsw, &DesQ::FSWatcher::nodeCreated, [ = ] ( QString path ) mutable {
            mParsePath = path;

            if ( not reloadTimer->isActive() ) {
                reloadTimer->start( 250, Qt::PreciseTimer, this );
            }
        }
    );
    connect(
        fsw, &DesQ::FSWatcher::nodeChanged, [ = ] ( QString path ) mutable {
            mParsePath = path;

            if ( not reloadTimer->isActive() ) {
                reloadTimer->start( 250, Qt::PreciseTimer, this );
            }
        }
    );
    connect(
        fsw, &DesQ::FSWatcher::nodeDeleted, [ = ] ( QString path ) mutable {
            mParsePath = path;

            if ( not reloadTimer->isActive() ) {
                reloadTimer->start( 250, Qt::PreciseTimer, this );
            }
        }
    );

    for ( QString path: desktopPaths ) {
        if ( QFile::exists( path ) ) {
            fsw->addWatch( path, DesQ::FSWatcher::PathOnly );
        }
    }

    fsw->startWatch();

    connect(
        this, &QListView::clicked, this, [ = ] ( QModelIndex item ) {
            QString desktopName = item.data( Qt::UserRole + 3 ).toString();
            DesQ::DesktopFile appFile( desktopName );
            appFile.startApplication( {} );
            emit closeMenu();
        }
    );

    connect(
        this, &QListWidget::activated, this, [ = ] ( QModelIndex item ) {
            QString desktopName = item.data( Qt::UserRole + 3 ).toString();
            DesQ::DesktopFile appFile( desktopName );
            appFile.startApplication( {} );
            emit closeMenu();
        }
    );
}


void DesQApplicationsView::loadCategory( QString category ) {
    proxyModel->setFilterRole( Qt::UserRole + 2 );

    if ( category == "All Applications" ) {
        proxyModel->setFilterFixedString( "" );
    }

    else {
        proxyModel->setFilterFixedString( category );
    }

    setFocus();
}


void DesQApplicationsView::setFilter( QString filter ) {
    proxyModel->setFilterRole( Qt::UserRole + 1 );
    proxyModel->setFilterWildcard( filter );
}


QStringList DesQApplicationsView::availableCategories() {
    return mAvailableCategories;
}


void DesQApplicationsView::parseDesktops() {
    QString path = mParsePath;

    if ( path.count() and DesQ::Utils::isFile( path ) ) {
        if ( not path.endsWith( ".desktop" ) ) {
            return;
        }

        QString           app = DesQ::Utils::baseName( path );
        DesQ::DesktopFile desktop( app );

        if ( not desktop.isValid() ) {
            return;
        }

        if ( desktop.type() != DesQ::DesktopFile::Type::Application ) {
            return;
        }

        if ( desktop.visible() ) {
            desktops[ app ] = desktop;
        }

        else {
            desktops[ app ] = DesQ::DesktopFile();
        }

        /** We've corrected the desktop. Now fix the model */
        QStandardItem *item = items[ app ];
        item->setIcon( QIcon::fromTheme( desktop.icon(), QIcon::fromTheme( "application-x-desktop" ) ) );
        item->setText( desktop.name() );

        item->setData( desktop.name() + desktop.genericName() + desktop.description(), Qt::UserRole + 1 );
        item->setData( desktop.category(),                                             Qt::UserRole + 2 );
        item->setData( desktop.desktopName(),                                          Qt::UserRole + 3 );

        /** All done. Time to leave */
        return;
    }

    QStringList scanPaths;

    if ( path.count() and DesQ::Utils::isDir( path ) ) {
        scanPaths << path;
    }

    else {
        scanPaths = desktopPaths;
    }

    for ( QString path: desktopPaths ) {
        QDir appDir( path );
        appDir.setFilter( QDir::Files );
        appDir.setNameFilters( { "*.desktop" } );

        Q_FOREACH ( QString app, appDir.entryList() ) {
            /** This application has already been processed at a higher ranked directory */
            if ( desktops.contains( app ) ) {
                continue;
            }

            /** Get the DesktopFile */
            DesQ::DesktopFile desktopFile( app );

            if ( not desktopFile.isValid() ) {
                continue;
            }

            if ( desktopFile.type() != DesQ::DesktopFile::Type::Application ) {
                continue;
            }

            if ( desktopFile.visible() ) {
                desktops[ app ] = desktopFile;
            }

            else {
                desktops[ app ] = DesQ::DesktopFile();
            }
        }
    }

    /** Clear the model */
    appsModel->clear();

    /** Add each entry to the model */
    for ( QString app: desktops.keys() ) {
        DesQ::DesktopFile desktop = desktops[ app ];

        if ( not desktop.isValid() ) {
            continue;
        }

        if ( not mAvailableCategories.contains( desktop.category() ) ) {
            mAvailableCategories << desktop.category();
        }

        QStandardItem *item = new QStandardItem();

        item->setIcon( QIcon::fromTheme( desktop.icon(), QIcon::fromTheme( "application-x-desktop" ) ) );
        item->setText( desktop.name() );

        item->setData( desktop.name() + desktop.genericName() + desktop.description(), Qt::UserRole + 1 );
        item->setData( desktop.category(),                                             Qt::UserRole + 2 );
        item->setData( desktop.desktopName(),                                          Qt::UserRole + 3 );

        items[ app ] = item;
        appsModel->appendRow( item );
    }

    proxyModel->sort( 0 );
    mParsePath = QString();
}


void DesQApplicationsView::removeDesktops( QString path ) {
    if ( path.endsWith( "*.desktop" ) ) {
        QString app( DesQ::Utils::baseName( path ) );
        desktops.remove( app );
        QStandardItem *item = items.take( app );

        if ( item and item->row() > 0 ) {
            appsModel->removeRow( items[ app ]->row() );
            delete item;
        }

        /** We're done here */
        return;
    }

    /**
     * If an entire folder was removed, we're not going to work too hard
     * This is on the user/sysadmin. Folders should not be removed for no reason.
     * Parse evenrything.
     */
    parseDesktops();
}


void DesQApplicationsView::mouseReleaseEvent( QMouseEvent *mEvent ) {
    if ( not indexAt( mEvent->pos() ).isValid() ) {
        emit closeMenu();
    }

    QListView::mouseReleaseEvent( mEvent );
}


void DesQApplicationsView::paintEvent( QPaintEvent *pEvent ) {
    QPainter painter( viewport() );

    painter.setPen( Qt::gray );
    painter.drawLine( geometry().topRight(), geometry().bottomRight() );
    painter.end();

    QListView::paintEvent( pEvent );
}


void DesQApplicationsView::resizeEvent( QResizeEvent *rEvent ) {
    rEvent->accept();

    QSize vSize = viewport()->size();

    /* Items per row */
    float items = vSize.width() / (144 + spacing() );

    /* Minimum width of all items */
    float itemsWidth = items * 144;

    /* Empty space remaining */
    float empty = vSize.width() - itemsWidth;

    /* Extra space per item */
    float extra = (empty / items) - 5;

    /* New grid size */
    setGridSize( QSize( 144 + extra, 108 ) );
}


void DesQApplicationsView::keyPressEvent( QKeyEvent *kEvent ) {
    if ( kEvent->key() == Qt::Key_Slash ) {
        emit focusSearch();
        kEvent->accept();

        return;
    }

    QListView::keyPressEvent( kEvent );
}


void DesQApplicationsView::focusInEvent( QFocusEvent *fEvent ) {
    if ( not selectionModel()->hasSelection() ) {
        setCurrentIndex( model()->index( 0, 0 ) );
    }

    QListView::focusInEvent( fEvent );
}


void DesQApplicationsView::timerEvent( QTimerEvent *tEvent ) {
    if ( tEvent->timerId() == reloadTimer->timerId() ) {
        reloadTimer->stop();
        parseDesktops();

        tEvent->accept();
        return;
    }

    QListView::timerEvent( tEvent );
}


DesQFavoritesView::DesQFavoritesView( QWidget *parent ) : QListWidget( parent ) {
    setViewMode( QListView::ListMode );
    setIconSize( QSize( 32, 32 ) );
    setGridSize( QSize( 230, 36 ) );
    setFixedWidth( 250 );
    setFrameStyle( QFrame::NoFrame );
    viewport()->setAutoFillBackground( false );

    QSettings favApps( "DesQ", "Menu" );

    Q_FOREACH ( QString desktop, favApps.value( "FavoriteApps" ).toStringList() ) {
        DesQ::DesktopFile appFile( desktop );
        QString           name   = appFile.name();
        QString           icoStr = appFile.icon();
        QIcon             icon   = QIcon::fromTheme( icoStr );

        if ( icon.isNull() ) {
            if ( QFile::exists( icoStr ) ) {
                icon = QIcon( icoStr );
            }

            else {
                icon = QIcon::fromTheme( "application-x-executable" );
            }
        }

        QListWidgetItem *item = new QListWidgetItem( icon, name );
        item->setData( Qt::UserRole + 1, appFile.desktopFileUrl() );
        item->setSizeHint( QSize( 230, 36 ) );
        item->setTextAlignment( Qt::AlignVCenter | Qt::AlignLeft );

        addItem( item );
    }

    connect( this, &QListWidget::itemClicked, this, [ = ] ( QListWidgetItem *item ) {
                 QString desktopUrl = item->data( Qt::UserRole + 1 ).toString();
                 DesQ::DesktopFile appFile( desktopUrl );
                 appFile.startApplication( {} );

                 emit closeMenu();
             }
    );

    connect( this, &QListWidget::itemActivated, this, [ = ] ( QListWidgetItem *item ) {
                 QString desktopUrl = item->data( Qt::UserRole + 1 ).toString();
                 DesQ::DesktopFile appFile( desktopUrl );
                 appFile.startApplication( {} );

                 emit closeMenu();
             }
    );
}


void DesQFavoritesView::mouseReleaseEvent( QMouseEvent *mEvent ) {
    if ( not itemAt( mEvent->pos() ) ) {
        emit closeMenu();
    }

    QListWidget::mouseReleaseEvent( mEvent );
}


void DesQFavoritesView::keyPressEvent( QKeyEvent *kEvent ) {
    if ( kEvent->key() == Qt::Key_Slash ) {
        emit focusSearch();
        kEvent->accept();

        return;
    }

    QListWidget::keyPressEvent( kEvent );
}


void DesQFavoritesView::focusInEvent( QFocusEvent * ) {
    if ( not selectedItems().count() ) {
        setCurrentItem( item( 0 ) );
    }
}


void DesQFavoritesView::focusOutEvent( QFocusEvent * ) {
    selectionModel()->clearSelection();
}


QSize DesQAppsViewDelegate::sizeHint( const QStyleOptionViewItem& option, const QModelIndex& index ) const {
    QSize size( QStyledItemDelegate::sizeHint( option, index ) );

    /* Icon View */
    if ( option.decorationPosition == QStyleOptionViewItem::Top ) {
        // size.setWidth( static_cast<int>(FileSystemView::gridWidth) - 5 );
        return QSize( 128, 96 );
    }

    /* Tile View */
    else if ( option.decorationPosition == QStyleOptionViewItem::Left ) {
        // size.setWidth( static_cast<int>(FileSystemView::gridWidth) - 5 );
        return QSize( 256, 72 );
    }

    return size;
}
